@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2006-2013 PLT Design Inc.
+  Copyright (c) 2006-2018 PLT Design Inc.
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
@@ -29,6 +29,22 @@
 /* Used by vector-set-performance-stats!: */
 int scheme_jit_malloced;
 
+void scheme_init_jit()
+{
+#ifdef CAN_INLINE_ALLOC
+  intptr_t max_alloc;
+  max_alloc = GC_max_nursery_object_size();
+  if ((sizeof(Scheme_Vector) + (JIT_MAX_VECTOR_INLINE_SIZE - mzFLEX_DELTA) * sizeof(Scheme_Object *)) > max_alloc) {
+    scheme_log_abort("Misconfigured: inlined vector size is greater than maximum allowed by GC");
+    abort();
+  }
+  if ((sizeof(Scheme_Structure) + ((JIT_MAX_STRUCT_FIELD_INLINE_COUNT - mzFLEX_DELTA) * sizeof(Scheme_Object *))) > max_alloc) {
+    scheme_log_abort("Misconfigured: inlined struct size is greater than maximum allowed by GC");
+    abort();
+  }
+#endif
+}
+
 /*========================================================================*/
 /*                              JIT buffer                                */
 /*========================================================================*/
@@ -43,7 +59,7 @@ THREAD_LOCAL_DECL(static intptr_t jit_buffer_cache_size);
 THREAD_LOCAL_DECL(static int jit_buffer_cache_registered);
 
 #ifdef SET_DEFAULT_LONG_JUMPS
-static int default_long_jumps;
+static volatile int default_long_jumps;
 static volatile uintptr_t code_low, code_high;
 #endif
 
@@ -58,7 +74,7 @@ int scheme_mz_retain_it(mz_jit_state *jitter, void *v)
     jitter->retain_start[jitter->retained] = v;
 #ifdef JIT_PRECISE_GC
     /* We just change an array that is marked indirectly for GC
-       via a Scheme_Native_Closure_Data. Write to that record
+       via a Scheme_Native_Lambda. Write to that record
        so that a minor GC will trace it and therefore trace
        the reatined array: */
     if (jitter->retaining_data) {
@@ -162,6 +178,16 @@ static int check_long_mode(uintptr_t low, uintptr_t size)
 
   return 0;
 }
+
+int scheme_check_long_mode(int jitter_long_jumps_default)
+{
+  /* In places are enabled, then we're relying on TSO here. If another
+     place provides a shared code pointer, then seeing that pointer
+     means that this place also sees any change to
+     `default_long_jumps` that was made before the pointer was
+     shared. */
+  return (jitter_long_jumps_default != default_long_jumps);
+}
 #endif
 
 
@@ -170,11 +196,21 @@ void *scheme_generate_one(mz_jit_state *old_jitter,
 			  void *data,
 			  int gcable,
 			  void *save_ptr,
-			  Scheme_Native_Closure_Data *ndata)
+			  Scheme_Native_Lambda *ndata)
+/* The given generate() function is called at least twice: once to gather
+   the size of the generated code (at a temporary location), and again
+   to generate the final code at its final location. The size of the
+   generated code must not depend on the deistation address. The
+   `retain_start' field of hte jitter record passed to generate() will
+   be NULL for a sizing run and non-NULL for a generation run.
+
+   In the unlikely event that a 64-bit build switches from 32-bit
+   branches to 64-bit branches, generate() might be called an extra
+   time in either mode. */
 {
   mz_jit_state _jitter;
   mz_jit_state *jitter = &_jitter;
-  void *buffer;
+  void *buffer, *prev_buffer = NULL;
   int mappings_buffer[JIT_INIT_MAPPINGS_SIZE];
   int *mappings = mappings_buffer;
   intptr_t size = JIT_BUFFER_INIT_SIZE, known_size = 0;
@@ -279,9 +315,8 @@ void *scheme_generate_one(mz_jit_state *old_jitter,
          then switch over to long-jump mode. */
       if (check_long_mode((uintptr_t)buffer, size)) {
         /* start over */
-        known_size = 0;
-        use_long_jumps = 1;
-        continue;
+        return scheme_generate_one(old_jitter, generate, data, gcable,
+                                   save_ptr, ndata);
       }
     }
 #endif
@@ -332,9 +367,8 @@ void *scheme_generate_one(mz_jit_state *old_jitter,
     if (!use_long_jumps) {
       if (check_long_mode((uintptr_t)buffer, size)) {
         /* start over */
-        known_size = 0;
-        use_long_jumps = 1;
-        continue;
+        return scheme_generate_one(old_jitter, generate, data, gcable,
+                                   save_ptr, ndata);
       }
     }
 #endif
@@ -351,10 +385,35 @@ void *scheme_generate_one(mz_jit_state *old_jitter,
     jitter->limit = (char *)jitter->limit + padding;
     if (PAST_LIMIT() || (jitter->retain_start
 			 && (jitter->retained > num_retained))) {
-      scheme_console_printf("JIT buffer overflow: %p [%p,%p] (%d)!!\n", 
+      scheme_console_printf("internal error in JIT;\n"
+                            " ending address %p not in [%p,%p] (%d)\n",
 			    jit_get_ip(), 
 			    buffer, jitter->limit,
 			    !!jitter->retain_start);
+      if (jitter->retain_start) {
+        const char *bp = buffer, *tend;
+        scheme_console_printf("  buffer content: {\n");
+        while (bp < jitter->limit) {
+          int d = 16;
+          while (d-- && (bp < jitter->limit)) {
+            scheme_console_printf("%d,", *(unsigned char *)(bp++));
+          }
+          scheme_console_printf("\n");
+        }
+        scheme_console_printf("}\n");
+        bp = (char *)prev_buffer;
+        tend = bp + ((char *)jitter->limit - (char *)buffer);
+        scheme_console_printf("  temporary buffer content: {\n");
+        while (bp < tend) {
+          int d = 16;
+          while (d-- && (bp < tend)) {
+            scheme_console_printf("%d,", *(unsigned char *)(bp++));
+          }
+          scheme_console_printf("\n");
+        }
+        scheme_console_printf("}\n");
+      }
+      scheme_log_abort("internal error: JIT buffer overflow");
       abort();
     }
 
@@ -398,6 +457,7 @@ void *scheme_generate_one(mz_jit_state *old_jitter,
 	  jit_buffer_cache = buffer;
 	  jit_buffer_cache_size = size;
 	}
+        prev_buffer = buffer;
       }
       /* looping to try again... */
     } else {
@@ -443,12 +503,11 @@ static void new_mapping(mz_jit_state *jitter)
   jitter->mappings[jitter->num_mappings] = 0;
 }
 
-void scheme_mz_pushr_p_it(mz_jit_state *jitter, int reg) 
-/* de-sync's rs */
+void scheme_extra_pushed(mz_jit_state *jitter, int n) 
 {
   int v;
 
-  jitter->extra_pushed++;
+  jitter->extra_pushed += n;
   if (jitter->extra_pushed > jitter->max_extra_pushed)
     jitter->max_extra_pushed = jitter->extra_pushed;
 
@@ -458,8 +517,14 @@ void scheme_mz_pushr_p_it(mz_jit_state *jitter, int reg)
     new_mapping(jitter);
   }
   v = (jitter->mappings[jitter->num_mappings]) >> 2;
-  v++;
-  jitter->mappings[jitter->num_mappings] = ((v << 2) | 0x1);
+  v += n;
+  jitter->mappings[jitter->num_mappings] = (((unsigned)v << 2) | 0x1);
+}
+
+void scheme_mz_pushr_p_it(mz_jit_state *jitter, int reg) 
+/* de-sync's rs */
+{
+  scheme_extra_pushed(jitter, 1);
   
   mz_rs_dec(1);
   CHECK_RUNSTACK_OVERFLOW_NOCL();
@@ -468,21 +533,30 @@ void scheme_mz_pushr_p_it(mz_jit_state *jitter, int reg)
   jitter->need_set_rs = 1;
 }
 
-void scheme_mz_popr_p_it(mz_jit_state *jitter, int reg, int discard) 
+void scheme_extra_popped(mz_jit_state *jitter, int n)
 /* de-sync's rs */
 {
   int v;
 
-  jitter->extra_pushed--;
+  if (PAST_LIMIT()) return;
+
+  jitter->extra_pushed -= n;
 
   JIT_ASSERT(jitter->mappings[jitter->num_mappings] & 0x1);
   JIT_ASSERT(!(jitter->mappings[jitter->num_mappings] & 0x2));
   v = jitter->mappings[jitter->num_mappings] >> 2;
-  v--;
+  v -= n;
+  JIT_ASSERT(v >= 0);
   if (!v)
     --jitter->num_mappings;
   else
-    jitter->mappings[jitter->num_mappings] = ((v << 2) | 0x1);
+    jitter->mappings[jitter->num_mappings] = (((unsigned)v << 2) | 0x1);
+}
+
+void scheme_mz_popr_p_it(mz_jit_state *jitter, int reg, int discard) 
+/* de-sync's rs */
+{
+  scheme_extra_popped(jitter, 1);
 
   if (!discard)
     mz_rs_ldr(reg);
@@ -501,6 +575,9 @@ void scheme_mz_runstack_skipped(mz_jit_state *jitter, int n)
 {
   int v;
 
+  if (!n) return;
+  if (PAST_LIMIT()) return;
+
   if (!(jitter->mappings[jitter->num_mappings] & 0x1)
       || (jitter->mappings[jitter->num_mappings] & 0x2)
       || (jitter->mappings[jitter->num_mappings] > 0)) {
@@ -509,7 +586,7 @@ void scheme_mz_runstack_skipped(mz_jit_state *jitter, int n)
   v = (jitter->mappings[jitter->num_mappings]) >> 2;
   JIT_ASSERT(v <= 0);
   v -= n;
-  jitter->mappings[jitter->num_mappings] = ((v << 2) | 0x1);
+  jitter->mappings[jitter->num_mappings] = (((unsigned)v << 2) | 0x1);
   jitter->self_pos += n;
 }
 
@@ -517,6 +594,9 @@ void scheme_mz_runstack_unskipped(mz_jit_state *jitter, int n)
 {
   int v;
 
+  if (!n) return;
+  if (PAST_LIMIT()) return;
+
   JIT_ASSERT(jitter->mappings[jitter->num_mappings] & 0x1);
   JIT_ASSERT(!(jitter->mappings[jitter->num_mappings] & 0x2));
   v = (jitter->mappings[jitter->num_mappings]) >> 2;
@@ -525,7 +605,7 @@ void scheme_mz_runstack_unskipped(mz_jit_state *jitter, int n)
   if (!v)
     --jitter->num_mappings;
   else
-    jitter->mappings[jitter->num_mappings] = ((v << 2) | 0x1);
+    jitter->mappings[jitter->num_mappings] = (((unsigned)v << 2) | 0x1);
   jitter->self_pos -= n;
 }
 
@@ -539,7 +619,7 @@ void scheme_mz_runstack_pushed(mz_jit_state *jitter, int n)
       || (jitter->mappings[jitter->num_mappings] & 0x3)) {
     new_mapping(jitter);
   }
-  jitter->mappings[jitter->num_mappings] += (n << 2);
+  jitter->mappings[jitter->num_mappings] += ((unsigned)n << 2);
   jitter->need_set_rs = 1;
 }
 
@@ -550,7 +630,7 @@ void scheme_mz_runstack_closure_pushed(mz_jit_state *jitter, int a, int flags)
     jitter->max_depth = jitter->depth;
   jitter->self_pos += 1;
   new_mapping(jitter);
-  jitter->mappings[jitter->num_mappings] = (a << 4) | (flags << 2) | 0x2;
+  jitter->mappings[jitter->num_mappings] = ((unsigned)a << 4) | ((unsigned)flags << 2) | 0x2;
   jitter->need_set_rs = 1;
   /* closures are never popped; they go away due to returns or tail calls */
 }
@@ -563,7 +643,7 @@ void scheme_mz_runstack_flonum_pushed(mz_jit_state *jitter, int pos)
     jitter->max_depth = jitter->depth;
   jitter->self_pos += 1;
   new_mapping(jitter);
-  jitter->mappings[jitter->num_mappings] = (pos << 2) | 0x3;
+  jitter->mappings[jitter->num_mappings] = ((unsigned)pos << 2) | 0x3;
   jitter->need_set_rs = 1;
   /* flonums are never popped; they go away due to returns or tail calls */
 }
@@ -572,6 +652,9 @@ void scheme_mz_runstack_flonum_pushed(mz_jit_state *jitter, int pos)
 void scheme_mz_runstack_popped(mz_jit_state *jitter, int n)
 {
   int v;
+
+  if (PAST_LIMIT()) return;
+
   jitter->depth -= n;
   jitter->self_pos -= n;
 
@@ -584,7 +667,7 @@ void scheme_mz_runstack_popped(mz_jit_state *jitter, int n)
   if (!v)
     --jitter->num_mappings;
   else
-    jitter->mappings[jitter->num_mappings] = (v << 2);
+    jitter->mappings[jitter->num_mappings] = ((unsigned)v << 2);
   jitter->need_set_rs = 1;
 }
 
@@ -805,6 +888,7 @@ void scheme_mz_unbox_save(mz_jit_state *jitter, mz_jit_unbox_state *r)
   }
 
 void scheme_mz_unbox_restore(mz_jit_state *jitter, mz_jit_unbox_state *r)
+/* can be called multipel times for an `r` by generate_branch() */
 {
   jitter->unbox = r->unbox;
 #ifdef MZ_LONG_DOUBLE
